//+------------------------------------------------------------------+
//|                                              qpu_EMA-Rainbow.mq5 |
//|                                       Copyright 2012, MetaDriver |
//|                                            MetaDriver@rambler.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2012, MetaDriver"
#property link      "MetaDriver@rambler.ru"
#property version   "1.00"

#define  CountLine  512

#property indicator_chart_window
#property indicator_buffers CountLine
#property indicator_plots   CountLine
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum clRealType
  {
   cl_Float,// float
   cl_Double,// double
  };
enum eFutureMode
  {
   Back,    //   ()
   Forward, //  () 
  };
  
input eFutureMode FutureMode=Forward; //     ;-)
input bool        UseCpu=true;  //  CPU
input clRealType  cl_RType=cl_Double;   // OpelCL RealType
input int         FastPeriod=3;
input int         PeriodStep=2;
input color       FastColor=clrPaleGoldenrod;
input color       MiddleColor=clrLightSalmon;
input color       SlowColor=clrMaroon;
input color       LFastColor=clrPaleVioletRed;
input color       LMiddleColor=clrDarkTurquoise;
input color       LSlowColor=clrMaroon;
//--- 
struct SBuffer
  {
   double            B[];
  };
//--- indicator buffers
SBuffer       EmaBuffer[CountLine];
//--- Calc buffers  and other vars
float         gpu_float_Buffer[][CountLine];
double        gpu_double_Buffer[][CountLine];
int          cl_BufLength=1024*128;

color Colors[CountLine];
bool First=true;
//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
int OnInit()
  {
//--- indicator buffers mapping
   InitColors();
   for(int i=0;i<CountLine;i++)
     {
      InitBuffers(i);
     }
//---
   return(0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void InitColors()
  {
   int Half = CountLine/2;
   for(int i=0;i<Half;i++)
     {
      if(i%4==3)
         Colors[i]=MixColor(LSlowColor,LMiddleColor,i*1.0/Half);
      else
         Colors[i]=MixColor(SlowColor,MiddleColor,i*1.0/Half);
     }
   for(int i=Half;i<CountLine;i++)
     {
      if(i%4==3)
         Colors[i]=MixColor(LMiddleColor,LFastColor,(i-Half)*1.0/Half);
      else
         Colors[i]=MixColor(MiddleColor,FastColor,(i-Half)*1.0/Half);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
color MixColor(color A,color B,double mix)
  {
   int r = (int)MathRound((B     & 0xFF) * mix + (A     & 0xFF) * (1. - mix));
   int g = (int)MathRound((B>>8  & 0xFF) * mix + (A>>8  & 0xFF) * (1. - mix));
   int b = (int)MathRound((B>>16 & 0xFF) * mix + (A>>16 & 0xFF) * (1. - mix));
   return color(r|g<<8|b<<16);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void InitBuffers(int index)
  {
   SetIndexBuffer(index,EmaBuffer[index].B,INDICATOR_DATA);
   PlotIndexSetInteger(index,PLOT_DRAW_TYPE,DRAW_LINE);
//   PlotIndexSetInteger(index,PLOT_DRAW_TYPE,DRAW_LINE);
   PlotIndexSetInteger(index,PLOT_LINE_COLOR,Colors[index]);
   if(index%8==7) PlotIndexSetInteger(index,PLOT_LINE_WIDTH,1);
   if(FutureMode==Back)   ArraySetAsSeries(EmaBuffer[index].B,true);
  }
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,const int prev_calculated,
                const int begin,const double &price[])
  {
//---
   if(FutureMode==Back)      ArraySetAsSeries(price,true);
   int limit=rates_total-prev_calculated;
   if(limit==1) limit++;
   long  StartTime;   if(First) {StartTime=GetTickCount();}
   if(prev_calculated==0)
     {
      limit-=begin+1;
      for(int i=0;i<CountLine;i++) { EmaBuffer[i].B[limit]=price[limit]; }
      if(!UseCpu) { CalcEMAsAtGpu(limit,price); }  //      ... :)
     }

   if(limit!=0)
     {
      for(int i=0;i<CountLine;i++)
        {
         double p=2./(FastPeriod+PeriodStep*(CountLine-i)+1);
         for(int j=limit-1;j>=0;j--)
           {
            //         EmaBuffer[i].B[j] = price[j]+(0.0001*i-0.0032)*5;
            EmaBuffer[i].B[j]=price[j]*p+EmaBuffer[i].B[j+1]*(1.-p);
           }
        }
     }
   if(First)
     {
      Print("Calculate ",rates_total," bars at ",
            (UseCpu ? "CPU, " : "GPU (OpenCL), "),"time = ",GetTickCount()-StartTime," ms" );
      First=false;
     }
//--- return value of prev_calculated for next call
   return(rates_total);
  }
//+------------------------------------------------------------------+
void CalcEMAsAtGpu(int &limit,const double &price[])
  {
   ResetLastError();
   int cl_Ctx;
   int cl_Prg;
   int cl_Krn;
   int cl_Mem_In;
   int cl_Mem_Out;
   if((cl_Ctx=CLContextCreate())==INVALID_HANDLE)
     {
      Print("OpenCL not found. Using CPU for calculation.");
      return;
      //} else { Print("CLContextCreate() .. OK!");
     }
   if((cl_Prg=CLProgramCreate(cl_Ctx,Get_cl_Source()))==INVALID_HANDLE)
     {
      CLContextFree(cl_Ctx);
      Print("OpenCL program create failed. Using CPU for calculation.");
      return;
      //} else { Print("CLProgramCreate() .. OK!");
     }
   if((cl_Krn=CLKernelCreate(cl_Prg,"gpu_Ema"))==INVALID_HANDLE)
     {
      CLProgramFree(cl_Prg);   CLContextFree(cl_Ctx);
      Print("OpenCL kernel create failed. Using CPU for calculation.");
      Print("GetLastError returned .. ",GetLastError());
      switch(GetLastError())
        {
         case ERR_OPENCL_INVALID_HANDLE:
            Print("Err ==    program OpenCL");
            break;
         case ERR_INVALID_PARAMETER:
            Print("Err ==   ");
            break;
         case ERR_OPENCL_TOO_LONG_KERNEL_NAME:
            Print("Err ==     127 ");
            break;
         case ERR_OPENCL_KERNEL_CREATE:
            Print("Err ==       OpenCL.");
            break;
         default: Print("Err ==  .");
         }
      return;
      //} else { Print("CLKernelCreate() .. OK!");
     }
   if((cl_Mem_In=CLBufferCreate(cl_Ctx,
      cl_BufLength*((cl_RType==cl_Float)?sizeof(float):sizeof(double)),
      CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLKernelFree(cl_Krn);   CLProgramFree(cl_Prg);   CLContextFree(cl_Ctx);
      Print("OpenCL input buffer create failed. Using CPU for calculation.");
      return;
      //} else { Print("CLBufferCreate(..In_Buffer) .. OK!"); 
     }
   if((cl_Mem_Out=CLBufferCreate(cl_Ctx,
      cl_BufLength*CountLine*((cl_RType==cl_Float)?sizeof(float):sizeof(double)),
      CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLBufferFree(cl_Mem_In);  CLKernelFree(cl_Krn);  CLProgramFree(cl_Prg);  CLContextFree(cl_Ctx);
      Print("OpenCL output buffer create failed. Using CPU for calculation.");
      return;
      //} else { Print("CLBufferCreate(..Out_Buffer) .. OK!");
     }
   float fIn[],fOut[];  //   -
   double dIn[],dOut[];
   if(cl_RType==cl_Float)
     {
      ArrayResize(fIn,cl_BufLength);
      ArrayResize(fOut,cl_BufLength*CountLine);
     }
   else
     {
      ArrayResize(dIn,cl_BufLength);
      ArrayResize(dOut,cl_BufLength*CountLine);
     }
//---     OpenCL
   CLSetKernelArg(cl_Krn,0,cl_BufLength);
   CLSetKernelArgMem(cl_Krn,1,cl_Mem_In);
   CLSetKernelArgMem(cl_Krn,2,cl_Mem_Out);
   uint kOfs[1]={0};
   uint kWork[1]={CountLine};
//--- Main loop  ---
   while(limit/(cl_BufLength-1)>0)
     {
      if(cl_RType==cl_Float)
        {
         for(int i=0;i<CountLine;i++)
           {
            fOut[i]=(float)EmaBuffer[i].B[limit];  //    Ema
           }
         CLBufferWrite(cl_Mem_Out,fOut,0,CountLine*sizeof(float));  //   gpu
         for(int i=1;i<cl_BufLength;i++)
           {
            fIn[i]=(float)price[limit-i];                 //  
           }
         CLBufferWrite(cl_Mem_In,fIn);  //   gpu
        }
      else // (cl_RType==cl_Double)
        {
         for(int i=0;i<CountLine;i++)
           {
            dOut[i]=EmaBuffer[i].B[limit];  //    Ema
           }
         CLBufferWrite(cl_Mem_Out,dOut,0,CountLine*sizeof(double));  //   gpu
         for(int i=1;i<cl_BufLength;i++)
           {
            dIn[i]=price[limit-i];                 //  
           }
         CLBufferWrite(cl_Mem_In,dIn);  //   gpu
        }

      CLExecute(cl_Krn,1,kOfs,kWork);

      if(cl_RType==cl_Float)
        {
         CLBufferRead(cl_Mem_Out,fOut);
         for(int i=1,r=limit-1; i<cl_BufLength; i++,r--)
           {
            for(int j=0;j<CountLine;j++)
              {
               EmaBuffer[j].B[r]=fOut[i*CountLine+j];
              }
           }
        }
      else
        {
         CLBufferRead(cl_Mem_Out,dOut);
         for(int i=1,r=limit-1; i<cl_BufLength; i++,r--)
           {
            for(int j=0;j<CountLine;j++)
              {
               EmaBuffer[j].B[r]=dOut[i*CountLine+j];
              }
           }
        }

      limit-=(cl_BufLength-1);
     }
   Print("CLExecute() .. OK!  Hello OpenCL !!");
   CLBufferFree(cl_Mem_Out);
   CLBufferFree(cl_Mem_In);
   CLKernelFree(cl_Krn);
   CLProgramFree(cl_Prg);
   CLContextFree(cl_Ctx);

  }
string i2s(int x) {return IntegerToString(x);}
//---
string Get_cl_Source()
  {
   string cl_Source=
                    ((cl_RType==cl_Float)
                    ?
                    "#define  real  float                                 \r\n"
                    :
                    "#ifdef   KHR_DP_EXTENSION                            \r\n"
                    "#pragma  OPENCL EXTENSION cl_khr_fp64 : enable       \r\n"
                    "#else                                                \r\n"
                    "#pragma  OPENCL EXTENSION cl_amd_fp64 : enable       \r\n"
                    "#endif                                               \r\n"
                    "#define  real  double                                \r\n"
                    )+
                    "                                                     \r\n"
                    "__kernel void  gpu_Ema(                              \r\n"
                    "            int    CountCalc,                        \r\n"
                    "    __global real  *PriceBuffer,                     \r\n"
                    "    __global real  *EmaBuffer)                       \r\n"
                    "  {                                                  \r\n"
                    "   int  gID = get_global_id(0);                      \r\n"
                    //                      double p=2./(HiPeriod+PeriodStep*(CountLine-i)+1);
                    "   real p=2./("+i2s(FastPeriod)+"+"+i2s(PeriodStep)+"* \r\n"
                    "             ("+i2s(CountLine)+"-gID)+1);            \r\n"
                    "   for(int i=1; i<CountCalc; i++)                    \r\n"
                    "     {                                               \r\n"
                    "        EmaBuffer[gID+"+i2s(CountLine)+"*i] =        \r\n"
                    "          EmaBuffer[gID+"+i2s(CountLine)+"*          \r\n"
                    "                                (i-1)]*(1.-p)+       \r\n"
                    "          PriceBuffer[i]*p;                          \r\n"
                    "     }  // for i                                     \r\n"
                    "  }     // Ema()                                     \r\n";
   return cl_Source;
  }
//+------------------------------------------------------------------+
